import warnings
import numpy as np
from sklearn.utils.extmath import randomized_svd
from scipy.sparse.linalg import svds
import torch
import random
import torch.optim as optim
from scipy.stats import entropy


def set_seed(seed=0):
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(0)


def center_scale(arr):
    return (arr - arr.mean(axis=0)) / arr.std(axis=0)


def drop_zero_variability_columns(arr, tol=1e-8):
    """
    Drop columns for which its standard deviation is zero in any one of the arrays in arr_list.
    Parameters
    ----------
    arr: np.ndarray of shape (n_samples, n_features)
        Data matrix
    tol: float, default=1e-8
        Any number less than tol is considered as zero
    Returns
    -------
    np.ndarray where no column has zero standard deviation
    """
    bad_columns = set()
    curr_std = np.std(arr, axis=0)
    for col in np.nonzero(np.abs(curr_std) < tol)[0]:
        bad_columns.add(col)
    good_columns = [i for i in range(arr.shape[1]) if i not in bad_columns]
    return arr[:, good_columns]


def get_neighborhood_composition(knn_indices, labels, full_labels=None):
    """
    Compute the composition of neighbors for each sample.
    Parameters
    ----------
    knn_indices: np.ndarray of shape (n_samples, n_neighbors)
        Each row represents the knn of that sample
    labels: np.ndarray of shape (n_samples, )
        Cluster labels
    full_labels: np.ndarray of shape (n_total_samples, )
        Cluster labels for all field of views combined

    Returns
    -------
    comp: np.ndarray of shape (n_samples, n_neighbors)
        The composition (in proportion) of neighbors for each sample.
    """

    labels = list(labels)
    n, k = knn_indices.shape
    if full_labels is not None:
        unique_clusters = np.sort(np.unique(full_labels))
    else:
        unique_clusters = np.sort(np.unique(labels))
    n_clusters = len(unique_clusters)
    label_to_clust_idx = {label: i for i, label in enumerate(unique_clusters)}

    comp = np.zeros((n, n_clusters))
    for i, neighbors in enumerate(knn_indices):
        good_neighbors = [nb for nb in neighbors if nb != -1]
        for nb in good_neighbors:
            comp[i, label_to_clust_idx[labels[nb]]] += 1

    return (comp.T / comp.sum(axis=1)).T


def robust_svd(arr, n_components, randomized=False, n_runs=1):
    """
    Do deterministic or randomized SVD on arr.
    Parameters
    ----------
    arr: np.array
        The array to do SVD on
    n_components: int
        Number of SVD components
    randomized: bool, default=False
        Whether to run randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error
    Returns
    -------
    u, s, vh: np.array
        u @ np.diag(s) @ vh is the reconstruction of the original arr
    """
    if randomized:
        best_err = float('inf')
        u, s, vh = None, None, None
        for _ in range(n_runs):
            curr_u, curr_s, curr_vh = randomized_svd(arr,
                                                     n_components=n_components,
                                                     random_state=None)
            curr_err = np.sum((arr - curr_u @ np.diag(curr_s) @ curr_vh)**2)
            if curr_err < best_err:
                best_err = curr_err
                u, s, vh = curr_u, curr_s, curr_vh
        assert u is not None and s is not None and vh is not None
    else:
        if n_runs > 1:
            warnings.warn("Doing deterministic SVD, n_runs reset to one.")
        u, s, vh = svds(arr * 1.0,
                        k=n_components)  # svds can not handle integer values
    return u, s, vh


def svd_denoise(arr, n_components=20, randomized=False, n_runs=1):
    """
    Compute best rank-n_components approximation of arr by SVD.
    Parameters
    ----------
    arr: np.array of shape (n_samples, n_features)
        Data matrix
    n_components: int, default=20
        Number of components to keep
    randomized: bool, default=False
        Whether to use randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error
    Returns
    -------
    arr: array_like of shape (n_samples, n_features)
        Rank-n_comopnents approximation of the input arr.
    """
    if n_components is None:
        return arr
    u, s, vh = robust_svd(arr,
                          n_components=n_components,
                          randomized=randomized,
                          n_runs=n_runs)
    return u @ np.diag(s) @ vh


def svd_embedding(arr, n_components=20, randomized=False, n_runs=1):
    """
    Compute rank-n_components SVD embeddings of arr.
    Parameters
    ----------
    arr: np.array of shape (n_samples, n_features)
        Data matrix
    n_components: int, default=20
        Number of components to keep
    randomized: bool, default=False
        Whether to use randomized SVD
    n_runs: int, default=1
        Run multiple times and take the realization with the lowest Frobenious reconstruction error
    Returns
    -------
    embeddings: array_like of shape (n_samples, n_components)
        Rank-n_comopnents SVD embedding of arr.
    """
    if n_components is None:
        return arr
    u, s, vh = robust_svd(arr,
                          n_components=n_components,
                          randomized=randomized,
                          n_runs=n_runs)
    return u @ np.diag(s)


def get_optimizer_and_scheduler(parameters,
                                OptimizerAlg='Adam',
                                optimizer_kwargs=None,
                                SchedulerAlg='StepLR',
                                scheduler_kwargs=None):
    """
    Initialize  optimizer and scheduler for model training.
    """

    if SchedulerAlg is not None:
        SchedulerAlg = getattr(optim.lr_scheduler, SchedulerAlg,
                               optim.lr_scheduler.StepLR)
    OptimizerAlg = getattr(optim, OptimizerAlg, optim.Adam)

    if optimizer_kwargs:
        optimizer = OptimizerAlg(parameters, **optimizer_kwargs)
    else:
        optimizer = OptimizerAlg(parameters)

    if scheduler_kwargs:
        scheduler = SchedulerAlg(optimizer, **scheduler_kwargs)
    else:
        if SchedulerAlg:
            scheduler = SchedulerAlg(optimizer)
        else:
            scheduler = None

    return optimizer, scheduler


def cluster_refine(label,
                   label_ref,
                   entropy_threshold=0.75,
                   concen_threshold=1,
                   max_breaks=3,
                   size_lim=50):
    """
    Refine clustering results from CellSNAP embedding to make sure robustenss. If a cluster from CellSNAP result contains a high mixture of population defined by original input population identity (eg acquire from Leiden clustering on the feature similarity graph), this cluster will be futher refined to represent the best biological status.
    Parameters
    ----------
    label: array
        Contains the clustering numbers stored in anndata from scanpy Leiden clustering on the final CellSNAP embedding.
    label_ref: array
        Contains the cell population identity labels (either from initial Leiden clustering or user provided) which was used to calculate neighborhood composition etc.
    entropy_threshold: float
        Entropy threshold to decide whether to refine a certain cluster from CellSNAP result. Higher value means more sensitive to population mixing.
    concen_threshold: float
        Concentration threshold to decide whether to refine a certain cluster from CellSNAP result. Higher value means more sensitive to population mixing.
    max_breaks: int
        Maximum number of splits for refining a certain cluster from CellSNAP result. Higher value means more splits in a cluster is allowed.
    """

    label_out = label.copy()
    label_out.name = label_out.name + '-refined'
    label_out = label_out.astype(str)
    ll = np.unique(label)
    for l in ll:
        ref_l = label_ref[label == l]
        ref_l_freq = ref_l.value_counts()
        if entropy(ref_l_freq) > entropy_threshold:
            for i in np.arange(max_breaks - 1):
                bb = label[label_ref == ref_l_freq.index[i]]
                if entropy(bb.value_counts(
                )) < concen_threshold and ref_l_freq.iloc[i] >= size_lim:
                    label_out[(label == l) & (
                        label_ref == ref_l_freq.index[i])] = l + '-' + str(i)

    return label_out.astype('category')


def clean_cluster(label):
    """
    Helper function to clean up the cluster names from output returned by function 'cluster_refine()'.
    Produced cluster numbers will be sorted based on population size.
    Return new cleaned up cluster numbers.
    """

    ll = label.value_counts().index.to_list()
    i = 0
    dd = {}
    for l in ll:
        dd[l] = str(i)
        i = i + 1
    res = []
    for item in label:
        t = dd[item]
        res.append(t)
    return res



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv


class SNAP_GNN_LITE(nn.Module):
    """
    Model class. Used when implementing single-SNAP-GNN model. LITE model means only doing CellSNAP
    with feature and neighborhood information, no image morphology information used.
    For more detail technical description of the model please refer to the CellSNAP manuscript.
    """

    def __init__(self, input_dim, out_dim, gnn_latent_dim=32):
        super().__init__()
        self.fc = nn.Linear(in_features=input_dim, out_features=gnn_latent_dim)
        self.gnn_conv1 = GCNConv(gnn_latent_dim, gnn_latent_dim)
        self.gnn_conv2 = GCNConv(gnn_latent_dim, out_dim * 2)

    def encoder(self, x, edge_index):
        x = F.relu(self.fc(x))
        x = self.gnn_conv1(x, edge_index)
        return x

    def forward(self, x, edge_index):
        x = F.relu(self.encoder(x, edge_index))
        x = self.gnn_conv2(x, edge_index)
        return x




class SNAP_GNN_DUO(nn.Module):

    """
    Model class. Used when the full SNAP-GNN-duo model is used.
    For more detail technical description of the model please refer to the CellSNAP manuscript.
    """

    def __init__(self,
                 out_dim,
                 feature_input_dim,
                 cnn_input_dim,
                 gnn_latent_dim=33,
                 proj_dim=32,
                 fc_out_dim=33,
                 cnn_out_dim=11):
        super().__init__()
        self.fc = nn.Linear(in_features=feature_input_dim,
                            out_features=proj_dim)
        self.cnn_fc = nn.Linear(in_features=cnn_input_dim,
                                out_features=proj_dim)
        self.feat_conv1 = GCNConv(proj_dim, proj_dim)
        self.feat_conv2 = GCNConv(proj_dim, fc_out_dim)

        self.spat_conv1 = GCNConv(proj_dim, proj_dim)
        self.spat_conv2 = GCNConv(proj_dim, cnn_out_dim)

        self.proj1 = nn.Linear(in_features=fc_out_dim + cnn_out_dim,
                               out_features=gnn_latent_dim)
        self.proj2 = nn.Linear(in_features=gnn_latent_dim,
                               out_features=out_dim * 2)

    def feat_gnn_encoder(self, feat, feat_edge_index):
        feat = F.relu(self.fc(feat))
        feat = F.relu(self.feat_conv1(feat, feat_edge_index))
        feat = self.feat_conv2(feat, feat_edge_index)

        return feat

    def spat_gnn_encoder(self, spat, spat_edge_index):
        spat = F.relu(self.cnn_fc(spat))
        spat = F.relu(self.spat_conv1(spat, spat_edge_index))
        spat = self.spat_conv2(spat, spat_edge_index)

        return spat

    def encoder(self, feat, spat, feat_edge_index, spat_edge_index):
        x_feat = self.feat_gnn_encoder(feat, feat_edge_index)
        x_spat = self.spat_gnn_encoder(spat, spat_edge_index)
        x = torch.cat((x_feat, x_spat), dim=1)
        return x

    def forward(self, feat, spat, feat_edge_index, spat_edge_index):
        x = F.relu(self.encoder(feat, spat, feat_edge_index, spat_edge_index))
        x = self.proj1(x)
        x = F.relu(x)
        x = self.proj2(x)
        return x


class SNAP_CNN(nn.Module):

    """
    Model class. SNAP-CNN model used for extracting tissue morphology information from images.
    For more detail technical description of the model please refer to the CellSNAP manuscript.
    """

    def __init__(self, cnn_latent_dim, output_dim):
        super().__init__()
        self.conv1 = nn.Conv2d(2, 16, kernel_size=2, stride=2)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=3, stride=1)
        self.conv3 = nn.Conv2d(32, 64, kernel_size=4, stride=2)
        self.conv4 = nn.Conv2d(64, 128, kernel_size=2, stride=1)
        self.conv5 = nn.Conv2d(128, 256, kernel_size=2, stride=1)
        self.conv6 = nn.Conv2d(256, 512, kernel_size=2, stride=1)
        self.fc1 = nn.Linear(2048, 1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc3 = nn.Linear(512, cnn_latent_dim)
        self.fc4 = nn.Linear(cnn_latent_dim, output_dim)

    def cnn_encoder(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = self.pool(F.relu(self.conv3(x)))
        x = self.pool(F.relu(self.conv4(x)))
        x = self.pool(F.relu(self.conv5(x)))
        x = F.relu(self.conv6(x))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def forward(self, x):
        x = self.cnn_encoder(x)
        x = self.fc4(F.relu(x))
        return x



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import scipy
import skimage
import sys

sys.path.append("../src/CellSNAP/")
from .utils import *
import os
from tqdm import tqdm
from skimage.io import imread
from .preprocessing import *
from .datasets import *
from .cellsnap import *

#############################
# example code in this file #
#############################


def main():
    # pipeline for codex murine dataset
    df = pd.read_csv('data/codex_murine/features_and_metadata.csv',
                     index_col=0)
    features_list = [
        'CD45', 'Ly6C', 'TCR', 'Ly6G', 'CD19', 'CD169', 'CD106', 'CD3',
        'CD1632', 'CD8a', 'CD90', 'F480', 'CD11c', 'Ter119', 'CD11b', 'IgD',
        'CD27', 'CD5', 'CD79b', 'CD71', 'CD31', 'CD4', 'IgM', 'B220', 'ERTR7',
        'MHCII', 'CD35', 'CD2135', 'CD44', 'nucl', 'NKp46'
    ]
    murine_dataset = SNAP_Dataset(
        df,
        features_list=features_list,
        nbhd_composition=15,
        feature_neighbor=15,
        spatial_neighbor=15,
        path2img='../../data/tutorial/codex_murine/processed_images')
    # prepare meta data
    murine_dataset.initialize(cent_x="centroid_x",
                              cent_y="centroid_y",
                              celltype="feature_labels",
                              pca_components=25,
                              cluster_res=1.0)

    # train CellSNAP
    device = 'cuda:1' if torch.cuda.is_available() else 'cpu'
    murine_cellsnap = CellSNAP(murine_dataset,
                               device,
                               cnn_model=True,
                               cnn_latent_dim=128,
                               gnn_latent_dim=32)
    # Load pretrained SNAP-CNN embedding
    murine_cellsnap.cnn_embedding = np.load(
        'data/codex_murine/results/SNAP_CNN_embedding.npy')
    murine_cellsnap.get_snap_embedding(round=3,
                                       k=32,
                                       learning_rate=1e-3,
                                       n_epochs=5000,
                                       loss_fn='MSELoss',
                                       OptimizerAlg='Adam',
                                       optimizer_kwargs={},
                                       SchedulerAlg=None,
                                       scheduler_kwargs={},
                                       verbose=True)
    # clustering and visualization
    murine_cellsnap.get_snap_clustering(neighbor=15, resolution=1.0)
    murine_cellsnap.visualize_umap(murine_cellsnap.snap_embedding,
                                   murine_cellsnap.snap_clustering)


if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import scipy
from skimage.measure import regionprops
from scipy.io import loadmat
from skimage.io import imread
from skimage.transform import resize
import os
from tqdm import tqdm
import matplotlib.pyplot as plt

def process_save_images(images,
                        locations,
                        size,
                        save_folder,
                        truncation,
                        power=5,
                        aggr=[[0], [1]],
                        pad=1000,
                        verbose=False):
    
    """
    Helper function to produce cropped images for each individual cells and same them out.
    The save out images will be used in the SNAP-CNN process to extract morphology encoding.
    Parameters
    ----------
    image: np.array with (H,W,2)
        One tissue image file in the format of numpy array. Note the supplied x, y location of cells from
        the initial df should be the same as their pixel location in this supplied image. Two channels
        corresponds to the membrane and nuclear channels.
    size: int
        Size of the cropped individual images for each cell.
    truncation: float
        Quantile value as threshold to binarize the input image.
    aggr: list
        Default not used
    pad: int
        Padding value around input image.
    """

    img_idx = 0
    image = images
    pad_image = np.zeros(
        (image.shape[0] + 2 * pad, image.shape[1] + 2 * pad, image.shape[2]))
    pad_image[pad:image.shape[0] + pad, pad:image.shape[1] + pad, :] = image
    truncate = np.quantile(pad_image, q=truncation, axis=(0, 1))
    truncate = truncate[None, None, :]

    pad_image[pad_image <= truncate] = 0
    pad_image[pad_image > truncate] = 1

    pad_image_sum = np.zeros([pad_image.shape[0], pad_image.shape[1], 2])
    pad_image_sum[:, :, 0] = np.sum(pad_image[:, :, i] for i in aggr[0])
    pad_image_sum[:, :, 1] = np.sum(pad_image[:, :, i] for i in aggr[1])

    sub_location = locations
    n_cells = sub_location.shape[0]
    power = len(str(n_cells))

    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    if verbose:
        print("Processing each cell...and saving!", flush=True)
    for i in tqdm(range(n_cells)):
        # process each cell
        center_x = sub_location[i][0]
        center_y = sub_location[i][1]
        cur_image = np.transpose(
            pad_image_sum[(int(center_x) - size // 2 + pad):(int(center_x) +
                                                             size // 2 + pad),
                          (int(center_y) - size // 2 +
                           pad):(int(center_y) + size // 2 + pad), :],
            (2, 0, 1)).astype(np.int8)
        assert (cur_image.shape == (2, size, size))
        if verbose:
            if i % 10000 == 1:
                plt.imshow(cur_image[0, :, :])
                plt.show()
                plt.imshow(cur_image[1, :, :])
                plt.show()

        np.save(file=os.path.join(save_folder, f"img_{i:0{power}d}"),
                arr=cur_image)
        img_idx += 1

    return


########################################################################################
#         Below is legacy functions not used in the actual CellSNAP pipeline.          #
########################################################################################

def get_cell_idx_partition(df):
    """
    Return partition s.t. partition[i] = [start_i, end_i] are the
    starting and ending indices (closed interval) in df for the i-th FOV (indexing from 1).
    """
    starts = [0]
    for i, curr_cell_idx in enumerate(df['cellLabelInImage']):
        if i > 0 and curr_cell_idx < df['cellLabelInImage'].iloc[i - 1]:
            starts.append(i)
    partition = []
    for i, s in enumerate(starts):
        if i < len(starts) - 1:
            partition.append([s, starts[i + 1] - 1])
        else:
            partition.append([s, df.shape[0] - 1])
    return partition


def add_cell_locations(df,
                       path_to_segmentation,
                       shape_of_views=(9, 7),
                       shape_of_each_view=(1008, 1344),
                       verbose=True):
    """
    Add three new columns to df: location coordinates (x, y) as well as which FOV is each cell in.
    """
    # get the splitting points of different FOVs
    partition = get_cell_idx_partition(df)
    centroid_x = []
    centroid_y = []
    cell_views = []
    for view_j in range(shape_of_views[1]):
        for view_i in range(shape_of_views[0]):
            view = view_j * shape_of_views[0] + view_i + 1
            topleft = [
                view_i * shape_of_each_view[0], view_j * shape_of_each_view[1]
            ]
            if verbose:
                print("Now at field of view {}, top-left coordinate is {}".
                      format(view, topleft),
                      flush=True)
            seg = scipy.io.loadmat(
                '{}point{}/nuclCD45_1.6_mpp0.8/segmentationParams.mat'.format(
                    path_to_segmentation, view))['newLmod']
            # get unique labels, excluding zero
            unique_seg_labels = list(np.unique(seg.flatten()))[1:]
            # calculate centroids
            props = regionprops(seg)
            # unique labels should align with props
            assert len(unique_seg_labels) == len(props)
            # build dict of seg_label: x and seg_label: y
            seg_label_to_x = {}
            seg_label_to_y = {}
            for i in range(len(props)):
                seg_label_to_x[unique_seg_labels[
                    i]] = props[i]['centroid'][0] + topleft[0]
                seg_label_to_y[unique_seg_labels[
                    i]] = props[i]['centroid'][1] + topleft[1]
            # fill the centroids of this segment of df
            start, end = partition[view - 1]
            for i in range(start, end + 1):
                centroid_x.append(
                    seg_label_to_x[df.iloc[i]['cellLabelInImage']])
                centroid_y.append(
                    seg_label_to_y[df.iloc[i]['cellLabelInImage']])
                cell_views.append(view)
    # add new columns
    df['centroid_x'] = centroid_x
    df['centroid_y'] = centroid_y
    df['field_of_view'] = cell_views


def select_useful_features(img, channels=('CD45', 'nucl')):
    """
    Select useful channels in the whole tiff image of shape (18, 1008, 1344, 3),
    and get an np array of shape (1008, 1344, 31).
    Finally return an np array of shape (1008, 1344, len(channels))
    """
    features = [
        'CD45', 'Ly6C', 'TCR', 'Ly6G', 'CD19', 'CD169', 'CD106', 'CD3',
        'CD1632', 'CD8a', 'CD90', 'F480', 'CD11c', 'Ter119', 'CD11b', 'IgD',
        'CD27', 'CD5', 'CD79b', 'CD71', 'CD31', 'CD4', 'IgM', 'B220', 'ERTR7',
        'MHCII', 'CD35', 'CD2135', 'CD44', 'nucl', 'NKp46'
    ]
    feature_to_idx = {f: i for i, f in enumerate(features)}
    indices = [feature_to_idx[f] for f in channels]
    img = np.concatenate([
        img[0, :, :, :],
        np.array([img[i, :, :, j] for i in range(1, 14)
                  for j in range(1, 3)]).transpose((1, 2, 0)),
        img[16, :, :, 2][:, :, np.newaxis], img[17, :, :, 1][:, :, np.newaxis]
    ],
                         axis=2)
    return img[:, :, indices]


import torch
from torch.utils.data import Dataset
from torchvision import transforms
import os
import numpy as np
import pandas as pd
from . import utils
from . import graph
from preprocessing import *


class SNAP_Dataset(Dataset):
    # SNAP Dataset Object

    def __init__(self,
                 df,
                 nbhd_composition=15,
                 feature_neighbor=15,
                 spatial_neighbor=15,
                 features_list=None,
                 path2img=None,
                 use_transform=False):
        """
        Form dataset of spatial single-cell data
        Parameters
        ----------

        df : pandas dataframe containing information:
            - feature expression values (eg protein or RNA values etc)
            - 2D cell location information (x,y):
                @ cell location need to be global in the case of multiple seperated FOV (eg TMAs) to avoid incorrect spatial
                adjacency and neighborhood information.
                @ if will be training SNAP-CNN (extracting morphology information from image files), the supplied 2D cell
                location inforamtion should be consistent with the pixel locations of cell location in the input image files. 
        nbhd_composition : int 
            Number of neighboring cells in neighborhood composition vector, default 15
        feature_neighbor : int
            Number of neighbors to consider in feature similarity graph
        spatial_neighbor : int
            Number of neighbors to consider in spatial similarity graph
        features_list : list(str)
            List of feature names to be used in CellSNAP process
        path2img : str, optional
            Path to images (to be saved, output from function '.prepare_images') if using images to extract morphological information (CNN step)
        use_transform : bool
            Indicate whether to use data augmentation if using images to extract morphological information (CNN step)

        """

        super().__init__()
        self.df = df
        self.nbhd_composition = nbhd_composition
        self.feature_neighbor = feature_neighbor
        self.spatial_neighbor = spatial_neighbor
        self.features_list = features_list
        self.path2img = path2img
        self.use_transform = use_transform
        if self.use_transform:
            self.transform = transforms.Compose([
                transforms.RandomRotation(degrees=180),
                transforms.RandomHorizontalFlip(),
                transforms.RandomVerticalFlip()
            ])

    def __len__(self):
        return self.labels.shape[0]

    def __getitem__(self, index):
        img = np.load(
            os.path.join(self.path2img, f"img_{index:0{self.power}d}.npy"))
        if self.use_transform:
            img = self.transform(torch.Tensor(img))
        labels = self.labels[index]
        return img, labels

    def initialize(self,
                   cent_x,
                   cent_y,
                   celltype='feature_labels',
                   pca_components=25,
                   cluster_res=1.0,
                   resolution_tol=0.05,
                   n_clusters=None,
                   n_runs=1):
        """
        Initialize the class object by calculating:
        -feature edges from similarity, used in SNAP-GNN
        -feature labels from feature values (population identity), default by Leiden clustering (skip if supplied)
        -neighborhood composition
        -spatial adjacency related information
        Parameters
        ----------

        cent_x, cent_y : str
            Column names of self.df, which contains location of each cell
        celltype : str
            Defualt to 'feature_labels' which will run leiden clustering clustering 
            and produce population identity of each cell. Otherwise set to a column in self.df 
            that contains the initial cell type annotation information
        pca_component : int
            Number of PCs used in expression graph. if 'None' no PCA reduction will be performed on the input feature profile
        cluster_res, resolution_tol : float, float
            Resolution parameter for leiden clustering on features
        cluster : int
            Leiden clustering parameter, fix the number of output labels
        n_runs : int
            Number of rounds for computing leiden clustering, default to be 1

        """

        # create metadata for the dataset
        self.n_cells = self.df.shape[0]
        self.power = len(str(self.n_cells))
        features = self.df[self.features_list].to_numpy(
        ) if self.features_list is not None else self.df.to_numpy()
        self.features = utils.center_scale(features)
        self.cluster_res = cluster_res
        self.resolution_tol = resolution_tol
        self.pca_components = pca_components

        # create feature edges and feature labels
        self.feature_edges = graph.get_feature_edges(
            arr=self.features,
            pca_components=self.pca_components,
            n_neighbors=self.feature_neighbor,
            metric='correlation',
            verbose=False)
        if celltype == "feature_labels":
            self.feature_labels = graph.graph_clustering(
                self.df.shape[0],
                self.feature_edges,
                resolution=self.cluster_res,
                n_clusters=n_clusters,
                n_runs=n_runs,
                resolution_tol=self.resolution_tol,
                seed=None,
                verbose=False)
            self.df['feature_labels'] = self.feature_labels

            print('Leiden clustering identified ' +
                  str(len(np.unique(self.feature_labels))) +
                  ' clusters as input population identity.')
        else:
            print('Predefined annotation identified ' +
                  str(len(np.unique(self.df[celltype]))) +
                  ' clusters as input population identity.')

        print('Calculating cell neighborhood composition matrix...')

        locations = self.df[[cent_x, cent_y]].to_numpy()
        self.locations = locations
        spatial_knn_indices = graph.get_spatial_knn_indices(
            locations=locations,
            n_neighbors=self.nbhd_composition,
            method='kd_tree')
        cell_nbhd = utils.get_neighborhood_composition(
            knn_indices=spatial_knn_indices,
            labels=self.df[celltype],
            full_labels=self.df[celltype])

        self.labels = cell_nbhd
        # create dual label for GNN
        cell_label = pd.get_dummies(self.feature_labels)
        dual_label = np.hstack([cell_label, self.labels])
        self.dual_labels = dual_label

        # create spatial edges
        self.spatial_edges = graph.get_spatial_edges(
            arr=locations, n_neighbors=self.spatial_neighbor, verbose=True)

    def prepare_images(self,
                       image,
                       size,
                       truncation,
                       aggr,
                       pad=1000,
                       verbose=False):
        """
        Helper function takes input of a whole tissue iamge with nuclear and membrane channel.
        Output individual cropped binarized images for each cell of their adjacent tissue information.
        The images will be saved in the pre-specified saving directory in function '.SNAP_Dataset'.
        Parameters
        ----------

        image: np.array with (H,W,2)
            One tissue image file in the format of numpy array. Note the supplied x, y location of cells from
            the initial df should be the same as their pixel location in this supplied image. two channels
            corresponds to the membrane and nuclear channels.
        size: int
            Size of the cropped individual images for each cell.
        truncation: float
            Quantile value as threshold to binarize the input image.
        aggr: list
            Default not used
        pad: int
            Padding value around input image.
        """

        print('Saving images...')
        process_save_images(images=image,
                            locations=self.locations,
                            size=size,
                            save_folder=self.path2img,
                            truncation=truncation,
                            power=self.power,
                            aggr=aggr,
                            pad=pad,
                            verbose=verbose)



import warnings
import numpy as np
import leidenalg
import igraph as ig
import scanpy as sc
import anndata as ad
from sklearn.neighbors import NearestNeighbors

from . import utils


def get_spatial_knn_indices(locations, n_neighbors=15, method='kd_tree'):
    """
    Compute k-nearest neighbors of locations.
    Parameters
    ----------
    locations: np.ndarray of shape (n_samples, 2)
        Data matrix
    n_neighbors: int
        Number of nearest neighbors
    method: str, default='kd_tree'
        Method to use when computing the nearest neighbors, one of ['ball_tree', 'kd_tree', 'brute']

    Returns
    -------
    knn_indices: np.ndarray of shape (n_samples, n_neighbors)
        Each row represents the knn of that sample
    """
    locations = np.array(locations)
    assert n_neighbors <= locations.shape[0]
    # k-NN indices, may be asymmetric
    _, knn_indices = NearestNeighbors(
        n_neighbors=n_neighbors,
        algorithm=method).fit(locations).kneighbors(locations)
    return knn_indices


def leiden_clustering(n,
                      edges,
                      resolution=1,
                      n_runs=1,
                      seed=None,
                      verbose=False):
    """
    Apply Leiden modularity maximization algorithm on the graph defined by edges and weights
    Parameters
    ----------
    n: int
        Number of edges in total
    edges: list of length two or three
        If length is two, then the graph is unweighted and each edge is (edges[0][i], edges[1][i]),
        if the length is three, then the graph is weighted and the weight of (edges[0][i], edges[1][i]) is edges[2][i].
    resolution: float, default=1
        Resolution parameter in Leiden algorithm
    n_runs: int, default=1
        Number of runs of Leiden algorithm, the run with the best modularity is taken as the output
    seed: None or int, default=None
        Random seed used. If None, use a random integer. If n_runs > 1, then seed will be reset to None.
    verbose: bool, default=True
        Whether to print progress
    Returns
    -------
    labels: np.array of shape (n,)
        Cluster labels
    """
    g = ig.Graph(directed=True)
    g.add_vertices(n)
    g.add_edges(list(zip(edges[0], edges[1])))
    if len(edges) > 2:
        g.es['weight'] = tuple(edges[2])

    if n_runs > 1 and seed is not None:
        seed = None
        warnings.warn('n_runs > 1, seed is reset to None.')

    partition_kwargs = {
        'n_iterations': -1,
        'seed': seed,
        'resolution_parameter': resolution
    }
    if len(edges) > 2:
        partition_kwargs['weights'] = np.array(g.es['weight']).astype(
            np.float64)

    partition_type = leidenalg.RBConfigurationVertexPartition

    best_modularity = float('-inf')
    best_labels = None
    if verbose and n_runs > 1:
        print('Running Leiden algorithm for {} times...'.format(n_runs),
              flush=True)
    for _ in range(n_runs):
        curr_part = leidenalg.find_partition(graph=g,
                                             partition_type=partition_type,
                                             **partition_kwargs)
        curr_modularity = curr_part.modularity
        if curr_modularity > best_modularity:
            best_modularity, best_labels = curr_modularity, np.array(
                curr_part.membership)

    assert best_labels is not None

    if verbose:
        if n_runs > 1:
            print('Best modularity among {} runs is {}.'.format(
                n_runs, best_modularity),
                  flush=True)
        else:
            print('Modularity is {}.'.format(best_modularity), flush=True)
        print('The label has {} distinct clusters.'.format(
            len(np.unique(best_labels))),
              flush=True)

    return best_labels


def graph_clustering(n,
                     edges,
                     resolution=1,
                     n_clusters=None,
                     n_runs=1,
                     resolution_tol=0.05,
                     seed=None,
                     verbose=False):
    """
    Cluster the graph defined by edges and weights using Leiden algorithm.
    Parameters
    ----------
    n: int
        Number of edges in total
    edges: list of length two or three
        If length is two, then the graph is unweighted and each edge is (edges[0][i], edges[1][i]),
        if the length is three, then the graph is weighted and the weight of (edges[0][i], edges[1][i]) is edges[2][i].
    resolution: None or int, default=1
        If not None, then this is the resolution parameter in the clustering algorithm,
        if None, then n_clusters must be not None.
    n_clusters: None or int, default=None
        If not None, use binary search to select the resolution parameter to achieve the desired number of clusters,
        if None, then resolution must be not None.
    n_runs: int, default=1
        Number of runs of Leiden algorithm, the run with the best modularity is taken as the output.
    resolution_tol: float, default=0.05
        Any resolution within the range of plus/minus resolution_tol will not be differentiated.
    seed: None or int, default=None
        Random seed used. If None, use a random integer. If n_runs > 1, then seed will be reset to None.
    verbose: bool, default=True
        Whether to print progress
    Returns
    -------
    labels: np.array of shape (n,)
        Cluster labels
    """
    assert (resolution is not None) ^ (n_clusters is not None)

    def cluster_func(res, vb):
        return leiden_clustering(n=n,
                                 edges=edges,
                                 resolution=res,
                                 n_runs=n_runs,
                                 seed=seed,
                                 verbose=vb)

    if resolution is not None:
        return cluster_func(resolution, verbose)

    right = 1
    while True:
        curr_labels = cluster_func(right, False)
        curr_n_clusters = len(np.unique(curr_labels))
        if verbose:
            print('A resolution of {} gives {} clusters.'.format(
                right, curr_n_clusters),
                  flush=True)
        if curr_n_clusters == n_clusters:
            return curr_labels
        elif curr_n_clusters < n_clusters:
            right *= 2
        else:
            break
    left = 0 if right == 1 else right / 2
    # desired resolution is in [left, right)
    while left + resolution_tol < right:
        mid = (left + right) / 2
        curr_labels = cluster_func(mid, False)
        curr_n_clusters = len(np.unique(curr_labels))
        if verbose:
            print('A resolution of {} gives {} clusters.'.format(
                mid, curr_n_clusters),
                  flush=True)
        if curr_n_clusters == n_clusters:
            return curr_labels
        elif curr_n_clusters > n_clusters:
            right = mid
        else:
            left = mid
    return curr_labels


def get_feature_edges(arr,
                      pca_components=None,
                      n_neighbors=15,
                      metric='correlation',
                      verbose=False):
    """
    Compute k-nearest neighbors of data and return the UMAP graph.
    Parameters
    ----------
    arr: np.array of shape (n_samples, n_features)
        Data matrix
    pca_components: None or int
        If None, then do not do PCA,
        else, number of components to keep when doing PCA de-noising for the data matrix.
    n_neighbors: int
        Number of neighbors desired
    metric: string, default='correlation'
        Metric used when constructing the initial knn graph
    verbose: bool, default=True
        Whether to print progress
    Returns
    -------
    rows, cols, vals: list
        Each edge is rows[i], cols[i], and its weight is vals[i]
    """
    arr = utils.drop_zero_variability_columns(arr=arr)
    if verbose:
        print("Constructing the graph...", flush=True)
    # use scanpy functions to do the graph construction
    adata = ad.AnnData(arr, dtype=np.float32)
    if pca_components is not None:
        sc.tl.pca(adata, n_comps=pca_components, svd_solver='arpack')
    if pca_components is not None:
        sc.pp.neighbors(adata,
                        n_neighbors=n_neighbors,
                        n_pcs=pca_components,
                        use_rep='X_pca',
                        metric=metric)
    else:
        sc.pp.neighbors(adata,
                        n_neighbors=n_neighbors,
                        n_pcs=None,
                        use_rep='X',
                        metric=metric)

    rows, cols = adata.obsp['connectivities'].nonzero()
    vals = adata.obsp['connectivities'][(rows, cols)].A1
    if verbose:
        print("Done!", flush=True)
    return rows, cols, vals


def get_spatial_edges(arr, n_neighbors=15, metric='euclidean', verbose=False):

    """
    Produce the spatial adjacency edges that will be used in SNAP-GNN-duo spatial-GNN part.
    Parameters
    ----------
    arr: np.array of shape (n_samples, 2)
        Array matrix that contains each cell's spatial location.
    n_neighbors: int
        Number of neighbors desired to be consider as spatially adjacent.
    metric: string, default='correlation'
        Metric used when constructing the initial knn graph
    verbose: bool, default=True
        Whether to print progress
    """

    if verbose:
        print("Constructing the graph...", flush=True)
    # use scanpy functions to do the graph construction
    adata = ad.AnnData(arr, dtype=np.float32)
    sc.pp.neighbors(adata,
                    n_neighbors=n_neighbors,
                    n_pcs=None,
                    use_rep='X',
                    metric=metric)

    rows, cols = adata.obsp['connectivities'].nonzero()
    vals = adata.obsp['connectivities'][(rows, cols)].A1
    if verbose:
        print("Done!", flush=True)
    return rows, cols, vals



import torch
from torch.utils.data import Dataset
from torchvision import transforms
import os
import numpy as np
import pandas as pd
from . import utils
from . import graph
from . import preprocessing
import anndata as ad
import scanpy as sc
from models import *
import torch.optim as optim
from matplotlib import pyplot as plt


class CellSNAP:

    def __init__(self,
                 dataset,
                 device,
                 cnn_model=None,
                 cnn_latent_dim=128,
                 gnn_latent_dim=33,
                 proj_dim=32,
                 fc_out_dim=33,
                 cnn_out_dim=11):
        """
        Initialize the CellSNAP object for model training.
        Parameters
        ----------

        dataset : SNAP_Dataset object
            Obtained from CellSNAP previous step (dataset preparation).
        device : str
            gpu or cpu device name.
        cnn_model : bool
            Default None, will implement CellSNAP without image morphology information.
        cnn_latent_dim : int
            Default 128. Latent dimension for SNAP-CNN model.
        gnn_latent_dim : int
            Default 33. Latent dimension for SNAP-GNN-duo model.
        proj_dim : int
            Default 32. Projection dimension for feature and CNN encoding inputs.
        fc_out_dim : int
            Default 33. Output dimension for expression-GNN and input to MLP head.
        cnn_out_dim : int
            Default 11. Output dimension for spatial-GNN and input to MLP head.
        """

        self.dataset = dataset
        self.device = device
        self.output_dim = self.dataset.labels.shape[1]
        self.n_cell = self.dataset.dual_labels.shape[0]
        self.gnn_latent_dim = gnn_latent_dim
        self.cnn_latent_dim = cnn_latent_dim
        self.embed_dim = fc_out_dim + cnn_out_dim if cnn_model else gnn_latent_dim
        self.cnn_model = cnn_model
        self.proj_dim = proj_dim
        self.fc_out_dim = fc_out_dim
        self.cnn_out_dim = cnn_out_dim

        return

    def fit_snap_cnn(self,
                     batch_size=64,
                     learning_rate=1e-4,
                     n_epochs=300,
                     loss_fn='MSELoss',
                     OptimizerAlg='Adam',
                     optimizer_kwargs=None,
                     SchedulerAlg=None,
                     scheduler_kwargs=None,
                     print_every=10):
        """
        Train SNAP-CNN to extract morphology encoding.
        Parameters
        ----------

        batch_size : int
            Batch size.
        learning_rate : float
            Learning rate.
        n_epochs : int
            Number of epochs to train.
        loss_fn : str
            Loss function name.
        OptimizerAlg : str
            Optimizer name.
        SchedulerAlg : None
        optimizer_kwargs : None
        print_every : int
            Log print frequency.
        """

        print(
            '\n=============Training convolutional neural network============\n',
            flush=True)
        self.cnn_model = SNAP_CNN(self.cnn_latent_dim, self.output_dim)
        # enable data augmentation
        self.dataset.use_transform = True
        dataloader = torch.utils.data.DataLoader(self.dataset,
                                                 batch_size=batch_size,
                                                 shuffle=True,
                                                 num_workers=1)
        criterion = getattr(nn, loss_fn, nn.MSELoss)()

        self.cnn_model = self.cnn_model.to(self.device)
        optimizer, scheduler = utils.get_optimizer_and_scheduler(
            self.cnn_model.parameters(), OptimizerAlg, {
                'lr': learning_rate,
                **optimizer_kwargs
            }, SchedulerAlg, scheduler_kwargs)
        criterion.to(self.device)
        self.cnn_model.train()
        for epoch in range(1, 1 +
                           n_epochs):  # loop over the dataset multiple times

            running_loss = 0.0
            running_sample = 0
            for i, data in enumerate(dataloader, 0):
                # get the inputs; data is a list of [inputs, labels]
                inputs, labels = data
                inputs = inputs.to(self.device).to(torch.float32)
                labels = labels.to(self.device).to(torch.float32)

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward + backward + optimize
                outputs = self.cnn_model(inputs)
                loss = criterion(outputs, labels)
                loss.backward()
                optimizer.step()
                # change learning rate
                if scheduler:
                    scheduler.step()

                running_loss += loss.item() * inputs.shape[0]
                running_sample += inputs.shape[0]
                if i % 100 == 99:
                    print(
                        f'===Epoch {epoch}, Step {i + 1:5d} loss: {running_loss / running_sample:.6f}==='
                    )

            if epoch % print_every == 0:
                print(
                    f'===Epoch {epoch}, Step {i + 1:5d} loss: {running_loss / running_sample:.6f}===',
                    flush=True)

        return

    def get_cnn_embedding(self, batch_size=512, path2result=None):
        """
        Retrieve SNAP-CNN embedding (extracted morphology encoding).
        Parameters
        ----------

        batch_size : int
            Batch size.
        path2result : str
            Directory to save output.

        """

        self.dataset.use_transform = False
        testloader = torch.utils.data.DataLoader(self.dataset,
                                                 batch_size=batch_size,
                                                 shuffle=False,
                                                 num_workers=1)
        self.cnn_model.to(self.device)
        self.cnn_model.eval()
        self.cnn_embedding = np.zeros(
            (self.dataset.df.shape[0], self.cnn_latent_dim))
        start_idx = 0
        with torch.no_grad():
            for data in testloader:
                inputs, labels = data
                inputs = inputs.to(self.device).to(torch.float32)
                outputs = self.cnn_model.cnn_encoder(inputs)
                self.cnn_embedding[start_idx:start_idx +
                                   inputs.shape[0]] = outputs.cpu()

                start_idx += inputs.shape[0]
        assert (start_idx == self.dataset.df.shape[0])
        print('\n=====Save CNN Embedding!============\n', flush=True)
        if path2result:
            if not os.path.exists(path2result):
                os.makedirs(path2result)
            np.save(os.path.join(path2result, "SNAP_CNN_embedding.npy"),
                    self.cnn_embedding)

    def fit_snap_gnn(self,
                     learning_rate=1e-4,
                     n_epochs=3000,
                     loss_fn='MSELoss',
                     OptimizerAlg='Adam',
                     optimizer_kwargs={},
                     SchedulerAlg=None,
                     scheduler_kwargs={},
                     print_every=500,
                     verbose=True):
        """
        Train SNAP-GNN-duo (or single SNAP-GNN model if morphology information is not provided).
        Parameters
        ----------

        batch_size : int
            Batch size.
        path2result : str
            Directory to save output.
        """

        if self.cnn_model:
            self.gnn_model = SNAP_GNN_DUO(
                out_dim=self.output_dim,
                feature_input_dim=self.dataset.features.shape[1],
                cnn_input_dim=self.cnn_latent_dim,
                gnn_latent_dim=self.gnn_latent_dim,
                proj_dim=self.proj_dim,
                fc_out_dim=self.fc_out_dim,
                cnn_out_dim=self.cnn_out_dim)
        else:
            self.gnn_model = SNAP_GNN_LITE(
                out_dim=self.output_dim,
                input_dim=self.dataset.features.shape[1],
                gnn_latent_dim=self.gnn_latent_dim)
        features = torch.from_numpy(self.dataset.features).float().to(
            self.device)
        features_edges = self.dataset.feature_edges
        spatial_edges = self.dataset.spatial_edges
        feat_edge_index = torch.from_numpy(np.array(
            features_edges[:2])).long().to(self.device)
        spatial_edge_index = torch.from_numpy(np.array(
            spatial_edges[:2])).long().to(self.device)
        cell_nbhd = torch.from_numpy(self.dataset.dual_labels).float().to(
            self.device)
        criterion = getattr(nn, loss_fn, nn.MSELoss)()
        criterion.to(self.device)
        self.gnn_model.to(self.device)
        optimizer, scheduler = utils.get_optimizer_and_scheduler(
            self.gnn_model.parameters(), OptimizerAlg, {
                'lr': learning_rate,
                **optimizer_kwargs
            }, SchedulerAlg, scheduler_kwargs)
        self.gnn_model.train()
        if self.cnn_model:
            cnn_embedding = torch.from_numpy(self.cnn_embedding).float().to(
                self.device)
        for e in range(1, 1 + n_epochs):
            if self.cnn_model:
                predicted_nbhd = self.gnn_model(
                    feat=features,
                    spat=cnn_embedding,
                    feat_edge_index=feat_edge_index,
                    spat_edge_index=spatial_edge_index)
            else:
                predicted_nbhd = self.gnn_model(x=features,
                                                edge_index=feat_edge_index)
            # Compute prediction error
            loss = criterion(predicted_nbhd, cell_nbhd)
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()

            # take one step
            optimizer.step()
            if scheduler:
                scheduler.step()

            # record the loss
            curr_train_loss = loss.item()
            if verbose and e % print_every == 0:
                print(
                    f'===Epoch {e}, the training loss is {curr_train_loss:>0.8f}==',
                    flush=True)

        print('\n=========Get Current Round CellSNAP Embedding!============\n',
              flush=True)
        self.gnn_model.eval()
        with torch.no_grad():
            if self.cnn_model:
                gnn_embedding = self.gnn_model.encoder(
                    feat=features,
                    spat=cnn_embedding,
                    feat_edge_index=feat_edge_index,
                    spat_edge_index=spatial_edge_index).detach().cpu().numpy()
            else:
                gnn_embedding = self.gnn_model.encoder(
                    x=features,
                    edge_index=feat_edge_index).detach().cpu().numpy()

        return gnn_embedding

    def get_snap_embedding(self,
                           round=5,
                           k=32,
                           learning_rate=1e-4,
                           n_epochs=3000,
                           loss_fn='MSELoss',
                           OptimizerAlg='Adam',
                           optimizer_kwargs=None,
                           SchedulerAlg=None,
                           scheduler_kwargs=None,
                           verbose=True):
        """
        Retrieve CellSNAP final embedding.
        Parameters
        ----------

        round: int
            Number of rounds the model the SNAP-GNN-duo model is trained to achieve robust performance.
        k: int
            Default to 32. Final out embedding dimension value.
        learning_rate : float
            Learning rate.
        n_epochs : int
            Number of epochs to train.
        loss_fn : str
            Loss function name.
        OptimizerAlg : str
            Optimizer name.
        SchedulerAlg : None
        optimizer_kwargs : None
        print_every : int
            Log print frequency.

        """
        dim = self.embed_dim
        concat_embedding = np.zeros((self.n_cell, round * dim))
        for i in range(round):
            gnn_embedding = self.fit_snap_gnn(
                learning_rate=learning_rate,
                n_epochs=n_epochs,
                loss_fn=loss_fn,
                OptimizerAlg=OptimizerAlg,
                optimizer_kwargs=optimizer_kwargs,
                SchedulerAlg=SchedulerAlg,
                scheduler_kwargs=scheduler_kwargs,
                print_every=500,
                verbose=verbose)
            concat_embedding[:, i * dim:(i + 1) * dim] = gnn_embedding
        Ue, Se, Vhe = np.linalg.svd(concat_embedding, full_matrices=False)
        gnn_embedding = Ue[:, :k] @ np.diag(Se[:k])
        self.snap_embedding = gnn_embedding
        return

    def fit_transform(self,
                      cnn_batch_size=64,
                      cnn_learning_rate=1e-4,
                      cnn_epochs=300,
                      cnn_loss_fn='MSELoss',
                      cnn_print=10,
                      round=5,
                      k=32,
                      gnn_learning_rate=1e-3,
                      gnn_epochs=3000,
                      gnn_loss_fn='MSELoss',
                      optim='Adam',
                      optim_kwargs=None,
                      sche=None,
                      sche_kwargs=None,
                      verbose=True,
                      path2result=None):
        """
        Wrapper function.
        """

        if self.cnn_model:
            self.fit_snap_cnn(self,
                              batch_size=cnn_batch_size,
                              learning_rate=cnn_learning_rate,
                              n_epochs=cnn_epochs,
                              loss_fn=cnn_loss_fn,
                              OptimizerAlg=optim,
                              optimizer_kwargs=optim_kwargs,
                              SchedulerAlg=sche,
                              scheduler_kwargs=sche_kwargs,
                              print_every=cnn_print)
            self.get_cnn_embedding(self,
                                   batch_size=512,
                                   path2result=path2result)

        self.get_snap_embedding(self,
                                round=round,
                                k=k,
                                learning_rate=gnn_learning_rate,
                                n_epochs=gnn_epochs,
                                loss_fn=gnn_loss_fn,
                                OptimizerAlg=optim,
                                optimizer_kwargs=optim_kwargs,
                                SchedulerAlg=sche,
                                scheduler_kwargs=sche_kwargs,
                                print_every=cnn_print,
                                verbose=verbose)

        return

    def get_snap_clustering(self,
                            neighbor=15,
                            resolution=1.0,
                            entropy_threshold=0.75,
                            concen_threshold=1,
                            max_breaks=3,
                            size_lim=50):

        ##### need update here.

        # compute cell clustering based on SNAP embedding
        feature_labels = self.dataset.feature_labels
        embedding = self.snap_embedding
        snap_adata = ad.AnnData(utils.drop_zero_variability_columns(embedding))
        snap_adata.obs['input'] = feature_labels
        sc.pp.scale(snap_adata)
        sc.pp.neighbors(snap_adata, n_neighbors=neighbor, use_rep='X')
        sc.tl.umap(snap_adata)
        sc.tl.leiden(snap_adata, resolution=resolution)
        # clean cluster
        from utils import cluster_refine, clean_cluster
        snap_refine = cluster_refine(label=snap_adata.obs['leiden'],
                                     label_ref=snap_adata.obs['input'],
                                     entropy_threshold=entropy_threshold,
                                     concen_threshold=concen_threshold,
                                     max_breaks=max_breaks,
                                     size_lim=size_lim)
        snap_clean = clean_cluster(snap_refine)
        self.snap_clustering = snap_clean
        return

    def visualize_umap(self, embedding, label):
        """
        Wrapped-up helper function to visualize (UMAP) the embedding.
        Using scanpy functions and default setup.
        Parameters
        ----------

        embedding : array
            Embedding as input
        label : array
            Contains the meta information you want to visualize on the UMAP plot.
        """

        adata = ad.AnnData(embedding)
        sc.pp.scale(adata)
        adata.obs['annotation'] = list(label)
        sc.tl.pca(adata, svd_solver='arpack')
        sc.pp.neighbors(adata, n_neighbors=10)
        sc.tl.umap(adata)
        plt.rcParams["figure.figsize"] = (10, 8)
        sc.pl.umap(adata, color='annotation', legend_fontsize=17, show=False)
        plt.show()
        return



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cellsnap",
    version="0.0.8",
    author="Sheng Gao, Bokai Zhu",
    author_email="gaosheng0321@gmail.com, zhubokai@mit.edu",
    description="A package for enhancing single-cell population delineation by integrating cross-domain information.",
    long_description="Official implementation of Cell Spatial And Neighborhood Pattern (CellSNAP), a computational method that learns a single-cell representation embedding by integrating cross-domain information from tissue samples. Through the analysis of datasets spanning spatial proteomic and spatial transcriptomic modalities, and across different tissue types and disease settings, we demonstrate CellSNAP’s capability to elucidate biologically relevant cell populations that were previously elusive due to the relinquished tissue morphological information from images",
    long_description_content_type="text/markdown",
    url="https://github.com/sggao/cellsnap",
    packages=setuptools.find_packages(),
    install_requires=['anndata>=0.10.6',
                        'leidenalg>=0.9.1',
                        'matplotlib>=3.7.1',
                        'numpy>=1.26.0',
                        'pandas>=2.0.0',
                        'python_igraph>=0.11.4',
                        'scanpy>=1.10.0',
                        'scikit_image>=0.21.0',
                        'scikit_learn>=1.2.2',
                        'scipy>=1.10.0',
                        'torch>=1.12.0',
                        'torch_geometric>=2.3.0',
                        'torchvision>=0.13.1',
                        'tqdm>=4.64.1'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)